package fsmNetworkExample;

/**
 * An example of a finite state machine composition.
 * 
 * A finite state machine that emulates "seconds" indicator on a clock,
 * i.e. its Moore output is a two-digit string representing a number 
 * from 00 to 59.
 * 
 * It accepts two input types:
 *  CLK_PULSE - moves the clock to the next state
 *  RESET     - resets the indicator to "00"
 *    
 * It generates no external Mealy output.
 * 
 * Internally, it is a composition of two FSMSimpleAlarmClocks.
 * The first one (fsmSecLo) counts from 0 to 9 and represents the second digit 
 * of the output.
 * The second one counts from 0 to 5 and represents the first digit
 * of the output.
 * 
 * The input is fed to the fsmSecLo.
 * The fsmSecLo's output (QUIET/BEEP) is fed as QUIET/CLK_PULSE to the fsmSecHi.
 * At the same time, when the fsmSecLo "beeps", it activates RESET signal
 * to fsmSecLo.
 * Similarly, when the fsmSecLo "beeps", it activates RESET signal
 * to both fsmSecLo and fsmSecHi.
 * 
 *                            ***** B L A C K  B O X **********************************************************
 *                            **                                                                             **
 *                            **               +----------------+                                            **
 * external pulse generator  ----------------> | in-CLK_PULSE   |            +----------------+              **
 *                            **               |  fsmSecLo   OUT| --+------> | in-CLK_PULSE   |              **
 * user button "reset" ------------+------+--> | in-RESET       |   |        |  fsmSecHi   OUT| -->---+      **
 *                            **   |      |    +----------------+   v   +--> | in-RESET       |       |      **
 *                            **   v      ^                         |   |    +----------------+       |      **
 *                            **   |      |                         |   |                             |      **
 *                            **   |      +------------<------------+   +---------------<-------------+      **                       
 *                            **   |                                    |                                    ** 
 *                            **   +----------------------->------------+                                    **
 *                            **                                                                             **
 *                            *********************************************************************************
 *
 *                        
 * The implementation has an important bug:
 * When the user requests "RESET", the value is reset immediately.
 * However, it has no impact on CLK_PULSE generation, so the delay between
 * "00" and "01" is quite arbitrary.
 * Fixing this bug is left as an exercise.
 * It should NOT be fixed by tweaking the external pulse generator - it is an external
 * signal source and the FSMClock should have no options how to modify it.
 * 
 * There are at least two ways how to fix it.
 * 
 * First:
 * In practice, the signal CLK_PULSE should be generated e.g. 100x per second.
 * It should be fed to a new FSMSimpleAlarmClock (say, fsmSecPulseGen) that would count
 * from 0 to 99 and generate "BEEP" once per second. Its output should be fed
 * to fsmSecLo as in the current implementation.
 * Indeed, this "fix" just moves the problem to a different scale - now the delay between
 * the states 0 and 1 of the fsmSecPulseGen is arbitrary. However, from the user's point
 * of view, the error is up to 1/100 s, which is negligible for a common clock.
 * 
 * Second:
 * The current implementation of fsmSecLo always increases its state until it reaches
 * the target state, i.e. it contains state transitions
 *   IN_CLK_PULSE:   (N)      --> (N+1)
 *                   (target) --> (target)
 * It could be modified so that the RESET signal generates a transition
 *   RESET:          (N)      --> (N')
 * where (N') is a new state with transitions defined as
 *   RESET:          (N')     --> (N')
 *   IN_CLK_PULSE    (N')     --> (0)
 * For example, for target=3 the FSM would be:
 *
 *                                                CLK_PULSE/
 *                                                QUIET
 *                                              +----<----+  
 *    CLK_PULSE/      CLK_PULSE/    CLK_PULSE/  |         |   
 *    QUIET           QUIET         BEEP        v         |          
 * 0  ----------> 1  ----------> 2  ----------> 3 --------+
 * ^              |              |              |
 * |              | RESET/       | RESET/       | RESET/
 * |              | QUIET        | QUIET        | QUIET
 * |              v              v              v
 * |              1'             2'             3'
 * |              |              |              |
 * |              | CLK_PULSE/   | CLK_PULSE/   | CLK_PULSE/
 * |              | QUIET        | QUIET        | QUIET
 * |              v              v              v
 * +-------<------+------<-------+------<-------+
 *
 * Now the FSM goes to the state 0 after CLK_PULSE signal (not directly after the RESET), 
 * i.e. no timing error is introduced.
 * This solution has its own drawback, too. In precise time measurement, e.g. in sports,
 * you should be able to tell the clock "start measuring NOW". This solution waits up to 
 * one second until the clock indicator is reset to 00.
 * 
 * So a real solution should combine both approaches.  
 * 
 */
public class FSMClock implements FSM {
	FSMSimpleAlarmClock fsmSecLo; 
	FSMSimpleAlarmClock fsmSecHi; 
	
	final public static int IN_CLK_PULSE = -1;
	final public static int IN_RESET = 1;
	final public static int OUT_QUIET = 0;

	public FSMClock() {
		fsmSecLo = new FSMSimpleAlarmClock(10);
		fsmSecHi = new FSMSimpleAlarmClock(6);
	}
	
	@Override
	public void reset() {
		fsmSecLo.reset();
		fsmSecHi.reset();		
	}

	@Override
	public int push(int input) {
		int out;

		if (input == IN_RESET) {
			reset();
		} else if (input == IN_CLK_PULSE) {
			out = fsmSecLo.push(FSMSimpleAlarmClock.IN_CLK_PULSE);
			if (out == FSMSimpleAlarmClock.OUT_BEEP) {
				out = fsmSecHi.push(FSMSimpleAlarmClock.IN_CLK_PULSE);
				fsmSecLo.reset();
				if (out == FSMSimpleAlarmClock.OUT_BEEP) {
					fsmSecHi.reset();
				}
			}
		}
		return OUT_QUIET;
	}

	@Override
	public String getStateOutput() {
		return fsmSecHi.getStateOutput() + fsmSecLo.getStateOutput();
	}

}
